import asyncio
import uuid
import re
from typing import Dict, Optional, Tuple, Union
from ..components.base import Component
from ..components.ui.element import Element
from ..components.ui.base.container import Container
from ..utils import RegexParser, RouteType


class PageMeta(type):
    _registry: Dict[str, Tuple["Page", int]] = {}
    _dynamic_routes: Dict[str, Tuple[re.Pattern, "Page", int]] = {}
    _regex_routes: Dict[re.Pattern, Tuple["Page", int]] = {}

    def __init__(cls, name, bases, attrs):
        if hasattr(cls, "router") and cls.router:
            priority = getattr(cls, "_priority", 0)
            cls._original_router = cls.router

            regex_pattern, route_type = RegexParser.compile_route(cls.router)
            cls._regex = regex_pattern
            cls._route_type = route_type

            if route_type in [RouteType.REGEX_PATTERN, RouteType.REGEX_STRING]:
                PageMeta._regex_routes[regex_pattern] = (cls, priority)
            elif route_type in [RouteType.DYNAMIC, RouteType.CATCH_ALL]:
                PageMeta._dynamic_routes[cls.router] = (regex_pattern, cls, priority)
            else:  # RouteType.STATIC
                PageMeta._registry[cls.router] = (cls, priority)

        super().__init__(name, bases, attrs)


class Page(metaclass=PageMeta):
    router: str = None
    _priority: int = 0
    _page_class_name: Optional[str] = None

    def __init__(self, params: Optional[Dict[str, str]] = None):
        self._component_instance_cache: Dict[str, Component] = {}
        self._rendered_component_keys: set[str] = set()
        self.params = params or {}

    def render(self, **params) -> Union[Element, Component]:
        raise NotImplementedError

    def get_page_class_name(self) -> str:
        if self._page_class_name is None:
            clean_router = RegexParser.get_clean_class_name(
                getattr(self, "_original_router", self.router)
            )
            self._page_class_name = (
                f"quillion-page-{clean_router}-{uuid.uuid4().hex[:6]}"
            )
        return self._page_class_name

    def _get_or_create_component_instance(
        self, new_component_declaration: Component
    ) -> Component:
        from ..core.app import Quillion

        key = new_component_declaration.key
        if not key:
            return new_component_declaration
        if key not in self._component_instance_cache:
            self._component_instance_cache[key] = new_component_declaration

            async def rerender_callback():
                app = Quillion._instance
                if app and app.websocket:
                    await app.render_current_page(app.websocket)

            new_component_declaration._rerender_callback = rerender_callback
        else:
            cached_instance = self._component_instance_cache[key]
            cached_instance.text = new_component_declaration.text
            cached_instance.inline_style_properties.update(
                new_component_declaration.inline_style_properties
            )
            for cls in new_component_declaration.css_classes:
                if cls not in cached_instance.css_classes:
                    cached_instance.css_classes.append(cls)
            if not cached_instance._rerender_callback:

                async def rerender_callback():
                    app = Quillion._instance
                    if app and app.websocket:
                        await app.render_current_page(app.websocket)

                cached_instance._rerender_callback = rerender_callback
            new_component_declaration = cached_instance
        self._rendered_component_keys.add(key)
        return new_component_declaration

    def _cleanup_old_component_instances(self):
        keys_to_remove = [
            k
            for k in self._component_instance_cache
            if k not in self._rendered_component_keys
        ]
        for key in keys_to_remove:
            self._component_instance_cache[key]._rerender_callback = None
            del self._component_instance_cache[key]
        self._rendered_component_keys.clear()



import asyncio
import inspect
from typing import Optional, Any, Callable
from pydantic import BaseModel, ValidationError


class StateMeta(type):
    def __init__(self, name, bases, attrs):
        super().__init__(name, bases, attrs)
        self._defaults = {}

        for key in list(attrs.keys()):
            if not key.startswith("_") and not (
                callable(attrs[key])
                or isinstance(attrs[key], (classmethod, staticmethod, property))
            ):
                self._defaults[key] = attrs[key]

        if self._defaults:
            try:
                validation_model = type(
                    f"{name}",
                    (BaseModel,),
                    {"__annotations__": attrs.get("__annotations__", {})},
                )
                validation_model(**self._defaults)
            except ValidationError as e:
                raise TypeError(f"Invalid default values in State class: {e}")

        for key in self._defaults:
            delattr(self, key)

    def __getattr__(self, name):
        return getattr(self.get_instance(), name)

    def get_instance(self):
        from ...core.app import Quillion

        app = Quillion._instance
        if app is None or app.websocket is None:
            raise RuntimeError("No active WebSocket connection for state access")
        if not hasattr(app, "_state_instances"):
            app._state_instances = {}
        if self not in app._state_instances:
            app._state_instances[self] = State(self)
        return app._state_instances[self]

    def set(self, **kwargs):
        instance = self.get_instance()
        for name, value in kwargs.items():
            if name in instance._data:
                if hasattr(self, "__annotations__") and name in self.__annotations__:
                    try:
                        validation_model = type(
                            f"ValidationModel",
                            (BaseModel,),
                            {"__annotations__": {name: self.__annotations__[name]}},
                        )
                        validation_model(**{name: value})
                    except ValidationError as e:
                        raise TypeError(
                            f"Invalid value for state variable '{name}': {e}"
                        )

                old_value = instance._data[name]
                instance._data[name] = value
                if old_value != value and instance._rerender_callback:
                    callback_result = instance._rerender_callback()
                    if inspect.iscoroutine(callback_result):
                        asyncio.create_task(callback_result)


class State(metaclass=StateMeta):
    _rerender_callback: Optional[Callable[[], Any]] = None

    def __init__(self, cls):
        self._cls = cls
        self._data = {k: v for k, v in cls._defaults.items()}

    def __getattr__(self, name):
        if name in self._data:
            return self._data[name]
        return super().__getattribute__(name)

    def _set_rerender_callback(self, callback: Callable[[], Any]):
        self._rerender_callback = callback



import asyncio
import inspect
from typing import Optional, Dict, Any, Callable, Tuple, List
from .state import State
from .ui.element import Element


class CSS:
    def __init__(self, files: List[str]):
        self.files = files

    def to_dict(self, app) -> Dict[str, Any]:
        return {
            "tag": "link",
            "attributes": {
                "rel": "stylesheet",
                "href": self.files[0] if self.files else "",
            },
            "children": [],
        }


class Component(Element):
    def __init__(self, tag: str = "div", **kwargs):
        super().__init__(tag=tag, **kwargs)
        self.props = kwargs
        self._hook_state: Dict[int, Any] = {}
        self._hook_index: int = 0
        self._rerender_callback: Optional[Callable[[], Any]] = None

    def _reset_hooks(self):
        self._hook_index = 0

    def _request_rerender(self):
        if self._rerender_callback:
            callback_result = self._rerender_callback()
            if not inspect.iscoroutine(callback_result):
                return
            asyncio.create_task(callback_result)

    def use_state(self, initial_value: Any) -> Tuple[Any, Callable[[Any], None]]:
        current_index = self._hook_index
        self._hook_index += 1
        if current_index not in self._hook_state:
            if inspect.isclass(initial_value) and issubclass(initial_value, State):
                state_instance = initial_value.get_instance()
                state_instance._set_rerender_callback(self._request_rerender)
                self._hook_state[current_index] = state_instance
            else:
                self._hook_state[current_index] = initial_value
        state_value = self._hook_state[current_index]
        if isinstance(state_value, State):
            return state_value
        else:

            def setter(new_value: Any):
                if callable(new_value):
                    self._hook_state[current_index] = new_value(
                        self._hook_state[current_index]
                    )
                else:
                    self._hook_state[current_index] = new_value
                self._request_rerender()

            return state_value, setter

    def render_component(self) -> Element:
        raise NotImplementedError

    def to_dict(self, app) -> Dict[str, Any]:
        self._reset_hooks()
        rendered_element = self.render_component()
        if self.key:
            rendered_element.key = self.key
        if self.css_classes:
            rendered_element.css_classes.extend(self.css_classes)
        if self.inline_style_properties:
            rendered_element.inline_style_properties.update(
                self.inline_style_properties
            )
        return rendered_element.to_dict(app)



import asyncio
import mimetypes
import os
from aiohttp import web


class AssetServer:
    def __init__(self, assets_dir: str = "/"):
        self.assets_dir = assets_dir
        self.app = web.Application()
        self.app.router.add_get("/{path:.*}", self.handle_request)

    async def handle_request(self, request: web.Request) -> web.Response:
        path = request.match_info["path"]
        file_path = os.path.join(self.assets_dir, path)

        if not os.path.abspath(file_path).startswith(os.path.abspath(self.assets_dir)):
            return web.HTTPForbidden()

        if not os.path.isfile(file_path):
            return web.HTTPNotFound()

        mime_type, _ = mimetypes.guess_type(file_path)
        if mime_type is None:
            mime_type = "application/octet-stream"

        return web.FileResponse(file_path, headers={"Content-Type": mime_type})

    def start(self, host: str = "0.0.0.0", port: int = 1338):
        loop = asyncio.get_event_loop()
        loop.create_task(web._run_app(self.app, host=host, port=port, print=None))



import base64
import os
import json
import websockets
from typing import Dict, Optional
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.backends import default_backend
from cryptography.exceptions import InvalidTag
from typing import Callable, Dict, List, Optional, Any, Type, TypeVar, Tuple


class Crypto:
    def __init__(self):
        self.client_x25519_private_keys: Dict[
            websockets.WebSocketServerProtocol, x25519.X25519PrivateKey
        ] = {}
        self.client_aes_keys: Dict[websockets.WebSocketServerProtocol, bytes] = {}

    async def handle_key_exchange(
        self, websocket: websockets.WebSocketServerProtocol, data: Dict[str, Any]
    ) -> bool:
        action = data.get("action")
        if action == "public_key":
            client_public_key_bytes = base64.b64decode(data.get("key"))
            server_private_key = x25519.X25519PrivateKey.generate()
            server_public_key = server_private_key.public_key()
            self.client_x25519_private_keys[websocket] = server_private_key
            shared_secret = server_private_key.exchange(
                x25519.X25519PublicKey.from_public_bytes(client_public_key_bytes)
            )
            hkdf = HKDF(
                algorithm=hashes.SHA256(),
                length=32,
                salt=None,
                info=b"quillion-aes-key",
                backend=default_backend(),
            )
            session_aes_key = hkdf.derive(shared_secret)
            self.client_aes_keys[websocket] = session_aes_key
            server_public_key_bytes = server_public_key.public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
            server_public_key_b64 = base64.b64encode(server_public_key_bytes).decode(
                "utf-8"
            )
            await websocket.send(
                json.dumps(
                    {
                        "action": "server_public_key",
                        "server_public_key": server_public_key_b64,
                    }
                )
            )
            return True
        else:
            return False

    async def decrypt_message(
        self, websocket: websockets.WebSocketServerProtocol, data: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        action = data.get("action")
        if action == "encrypted_message":
            encrypted_data_b64 = data.get("data")
            nonce_b64 = data.get("nonce")
            encrypted_data = base64.b64decode(encrypted_data_b64)
            nonce = base64.b64decode(nonce_b64)
            session_aes_key = self.client_aes_keys.get(websocket)
            aesgcm = AESGCM(session_aes_key)
            decrypted_payload_bytes = aesgcm.decrypt(nonce, encrypted_data, None)
            decrypted_payload_str = decrypted_payload_bytes.decode("utf-8")
            inner_data = json.loads(decrypted_payload_str)
            return inner_data
        else:
            print(
                f"[{websocket.remote_address[0]}:{websocket.remote_address[1]}] Unknown action: {action}."
            )
        return None

    def encrypt_response(
        self, websocket: websockets.WebSocketServerProtocol, content: Dict[str, Any]
    ) -> Dict[str, Any]:
        session_aes_key = self.client_aes_keys.get(websocket)
        plaintext = json.dumps(content).encode("utf-8")
        nonce = os.urandom(12)
        aesgcm = AESGCM(session_aes_key)
        ciphertext = aesgcm.encrypt(nonce, plaintext, None)
        encrypted_payload_b64 = base64.b64encode(ciphertext).decode("utf-8")
        nonce_b64 = base64.b64encode(nonce).decode("utf-8")
        return {
            "action": "encrypted_response",
            "encrypted_payload": encrypted_payload_b64,
            "nonce": nonce_b64,
        }

    def cleanup(self, websocket: websockets.WebSocketServerProtocol):
        if websocket in self.client_x25519_private_keys:
            del self.client_x25519_private_keys[websocket]
        if websocket in self.client_aes_keys:
            del self.client_aes_keys[websocket]



from typing import Callable, Union, Pattern
import inspect
from pydantic import validate_arguments


def page(route: Union[str, Pattern], priority: int = 0):
    from ..pages.base import Page, PageMeta

    def decorator(func: Callable):
        validated_func = validate_arguments(func)
        is_async = inspect.iscoroutinefunction(func)

        class GeneratedPage(Page, metaclass=PageMeta):
            router = route
            _priority = priority

            if is_async:

                async def render(self, **params):
                    return await validated_func(**params)

            else:

                def render(self, **params):
                    return validated_func(**params)

        GeneratedPage.__name__ = func.__name__
        return GeneratedPage

    return decorator



from typing import List
import nest_asyncio
from .core import *
from .pages import *
from . import components
from .utils import *

nest_asyncio.apply()

app = Quillion()


def css(files: List[str]):
    return app.css(files)



import re
from enum import Enum, auto
from typing import Dict, Tuple, Optional, Union, Pattern


class RouteType(Enum):
    REGEX_PATTERN = auto()  # re.Pattern
    REGEX_STRING = auto()  # regex:
    CATCH_ALL = auto()  # .* / *
    DYNAMIC = auto()  # {param} / *
    STATIC = auto()


class RegexParser:
    @staticmethod
    def compile_route(route: Union[str, Pattern]) -> Tuple[Pattern, RouteType]:
        if isinstance(route, Pattern):
            return route, RouteType.REGEX_PATTERN

        if isinstance(route, str) and route.startswith("regex:"):
            regex_pattern = route[6:].strip()
            return re.compile(regex_pattern), RouteType.REGEX_STRING

        if route in [".*", "*"]:
            return re.compile(r".*"), RouteType.CATCH_ALL

        if "{" in route or "*" in route:
            pattern = re.escape(route)
            pattern = re.sub(r"\\\{(\w+)\\\}", r"(?P<\1>[^\/]+)", pattern)
            pattern = pattern.replace(r"\*", r"[^\/]+")
            return re.compile(f"^{pattern}$"), RouteType.DYNAMIC

        pattern = re.escape(route)
        return re.compile(f"^{pattern}$"), RouteType.STATIC

    @staticmethod
    def get_route_type(route: Union[str, Pattern]) -> RouteType:
        if isinstance(route, Pattern):
            return RouteType.REGEX_PATTERN

        if isinstance(route, str) and route.startswith("regex:"):
            return RouteType.REGEX_STRING

        if route in [".*", "*"]:
            return RouteType.CATCH_ALL

        if "{" in route or "*" in route:
            return RouteType.DYNAMIC

        return RouteType.STATIC

    @staticmethod
    def extract_params(pattern: Pattern, path: str) -> Optional[Dict]:
        match = pattern.match(path)
        return match.groupdict() if match else None

    @staticmethod
    def get_clean_class_name(route: Union[str, Pattern]) -> str:
        if isinstance(route, Pattern):
            route_str = route.pattern
        else:
            route_str = str(route)

        clean_router = route_str.strip("/").replace("/", "-")

        clean_router = re.sub(r"[^a-zA-Z0-9_-]", "", clean_router)

        if not clean_router or clean_router == "regex":
            clean_router = "dynamic"

        return clean_router
